package net.antonioshome.clexec;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

/**
 * Responsible for setting up stuff for launching a process.
 * Creates pipes for each one of the three streams (stdin, stdout, stderr) and
 * launches the process. Cleans up stuff before exiting. Handles interruption.
 * @author Antonio Vieiro (antonio@antonioshome.net)
 * @author $Author: $
 * @version $Revision: $
 */
class ProcessWrapper
  implements Callable<Integer>
{
  private Process process;
  private ProcessStreams inputOutput;

  ProcessWrapper( ProcessStreams anInputOutput, Process aProcess )
  {
    this.process = aProcess;
    this.inputOutput = anInputOutput;
  }

  public Integer call()
  {

    // Prepare pipes
    InputStream2OutputStreamPipe outMonitor = 
      new InputStream2OutputStreamPipe( process.getInputStream(), inputOutput.getStdout(), inputOutput.getStderr() );
    InputStream2OutputStreamPipe errMonitor =
      new InputStream2OutputStreamPipe( process.getErrorStream(), inputOutput.getStderr(), inputOutput.getStderr() );
    InputStream2OutputStreamPipe inMonitor =
      new InputStream2OutputStreamPipe( inputOutput.getStdin(), process.getOutputStream(), inputOutput.getStderr() );

    // The executorService used for the pipes: the ProcessExecutionService
    ExecutorService pipeExecutorService = ProcessExecutionService.getService();

    // Launch pipes
    Future<?> outFuture = pipeExecutorService.submit( outMonitor );
    Future<?> errFuture = pipeExecutorService.submit( errMonitor );
    Future<?> inFuture = pipeExecutorService.submit( inMonitor );

    try
    {
      // Start the process and wait until finished
      int returnValue = process.waitFor();

      /* Wait quickly for all streams to send data. If they don't
       * send data in a reasonable time (a few milliseconds) then just
       * interrupt those threads. */
      kindlyInterrupt( outFuture );
      kindlyInterrupt( errFuture );
      kindlyInterrupt( inFuture );

      /* Close process streams */
      silentlyClose( process.getInputStream() );
      silentlyClose( process.getOutputStream() );
      silentlyClose( process.getErrorStream() );

      /* Destroy the process, just in case */
      process.destroy();

      return returnValue;
    }
    catch (InterruptedException ex)
    {
      /* Interrupt working pipes */
      outFuture.cancel( true );
      errFuture.cancel( true );
      inFuture.cancel( true );

      /* Close streams */
      silentlyClose( process.getInputStream() );
      silentlyClose( process.getOutputStream() );
      silentlyClose( process.getErrorStream() );

      /* Send interrupted exception to stderr */
      ex.printStackTrace( inputOutput.getStderr() );

      /* Destroy the process */
      process.destroy();

      /* Don't return 0: we were interrupted */
      return -1;
    }

  }

  /**
   * Silently closes an InputStream
   * @param anInputStream the InputStream.
   */
  private void silentlyClose( InputStream anInputStream  )
  {
    try
    {
      anInputStream.close();
    }
    catch( IOException ioe )
    {
    }
  }

  /**
   * Silently closes an OutputStream
   * @param anOutputStream the OutputStream
   */
  private void silentlyClose( OutputStream anOutputStream  )
  {
    try
    {
      anOutputStream.close();
    }
    catch( IOException ioe )
    {
      ioe.printStackTrace();
    }
  }

  /**
   * Waits 500 milliseconds for a future to end, if not then interrupts
   *   the future.
   * @param aFuture the future.
   */
  private void kindlyInterrupt( Future<?> aFuture )
  {
    try
    {
      // Wait a little bit just in case the future is not done yet...
      if ( ! aFuture.isDone() )
        aFuture.get(500, TimeUnit.MILLISECONDS);
    }
    catch (InterruptedException ex)
    {
    }
    catch (ExecutionException ex)
    {
    }
    catch (TimeoutException ex)
    {
      // ... and if the future does not cooperate then just force an interruption
      aFuture.cancel( true );
    }
  }
}
